from typing import List, TYPE_CHECKING

from efootprint.abstract_modeling_classes.explainable_object_dict import ExplainableObjectDict
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.sources import Sources
from efootprint.constants.units import u

if TYPE_CHECKING:
    from efootprint.core.usage.usage_pattern import UsagePattern
    from efootprint.core.usage.edge.edge_usage_pattern import EdgeUsagePattern
    from efootprint.core.usage.job import JobBase


class Network(ModelingObject):
    default_values = {
            "bandwidth_energy_intensity": SourceValue(0.1 * u.kWh / u.GB)
        }

    @classmethod
    def wifi_network(cls, name: str = "Default wifi network"):
        return cls(
            name=name, bandwidth_energy_intensity=SourceValue(0.05 * u("kWh/GB"), Sources.TRAFICOM_STUDY))

    @classmethod
    def mobile_network(cls, name: str = "Default mobile network"):
        return cls(
            name=name, bandwidth_energy_intensity=SourceValue(0.12 * u("kWh/GB"), Sources.TRAFICOM_STUDY))

    @classmethod
    def archetypes(cls):
        return [cls.wifi_network, cls.mobile_network]

    def __init__(self, name: str, bandwidth_energy_intensity: ExplainableQuantity):
        super().__init__(name)
        self.energy_footprint_per_job = ExplainableObjectDict()
        self.energy_footprint = EmptyExplainableObject()
        self.instances_fabrication_footprint = EmptyExplainableObject()
        self.bandwidth_energy_intensity = bandwidth_energy_intensity.set_label(
            f"bandwith energy intensity of {self.name}")

    @property
    def calculated_attributes(self):
        return [
            "energy_footprint_per_job",
            "instances_fabrication_footprint",
            "energy_footprint",
        ] + [
            attr for attr in super().calculated_attributes if attr != "usage_impact_repartition_weights"
        ]

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List:
        return []

    @property
    def usage_patterns(self) -> List["UsagePattern"] | List["EdgeUsagePattern"]:
        return self.modeling_obj_containers

    @property
    def jobs(self) -> List["JobBase"]:
        return list(dict.fromkeys(sum([up.jobs for up in self.usage_patterns], start=[])))

    def update_instances_fabrication_footprint(self):
        self.instances_fabrication_footprint = EmptyExplainableObject()

    def update_dict_element_in_energy_footprint_per_job(self, job: "JobBase"):
        energy_footprint = EmptyExplainableObject()
        for usage_pattern in [up for up in job.usage_patterns if up in self.usage_patterns]:
            energy_footprint += (
                self.bandwidth_energy_intensity
                * job.hourly_data_transferred_per_usage_pattern[usage_pattern]
            ).to(u.kWh) * usage_pattern.country.average_carbon_intensity

        self.energy_footprint_per_job[job] = energy_footprint.to(u.kg).set_label(
            f"{job.name} energy footprint in {self.name}"
        )

    def update_energy_footprint_per_job(self):
        self.energy_footprint_per_job = ExplainableObjectDict()
        for job in self.jobs:
            self.update_dict_element_in_energy_footprint_per_job(job)

    def update_energy_footprint(self):
        self.energy_footprint = sum(self.energy_footprint_per_job.values(), start=EmptyExplainableObject()).set_label(
            f"Hourly {self.name} energy footprint"
        )

    def update_dict_element_in_fabrication_impact_repartition_weights(self, job: "JobBase"):
        raise NotImplementedError(
            f"Fabrication impact repartition is not implemented for {self.name} when Network has a fabrication footprint."
        )

    def update_fabrication_impact_repartition_weights(self):
        self.fabrication_impact_repartition_weights = ExplainableObjectDict()
        if isinstance(self.instances_fabrication_footprint, EmptyExplainableObject):
            return
        for job in self.jobs:
            self.update_dict_element_in_fabrication_impact_repartition_weights(job)

    @property
    def usage_impact_repartition_weights(self):
        return self.energy_footprint_per_job



import math
from collections.abc import Sequence

import numpy as np
from pint import Quantity, Unit

from efootprint.constants.units import u

UNIT_FAMILIES: list[Sequence[Unit]] = [
    [u.mg, u.g, u.kg, u.tonne],
    [u.mWh, u.Wh, u.kWh, u.MWh, u.GWh],
    [u.mW, u.W, u.kW, u.MW, u.GW],
    [u.occurrence, u.koccurrence, u.Moccurrence, u.Goccurrence],
    [u.concurrent, u.kconcurrent, u.Mconcurrent, u.Gconcurrent],
    [u.byte, u.kB, u.MB, u.GB, u.TB],
    [u.byte_ram, u.kB_ram, u.MB_ram, u.GB_ram, u.TB_ram],
    [u.byte_stored, u.kB_stored, u.MB_stored, u.GB_stored, u.TB_stored],
]


def _get_unit_family(quantity: Quantity) -> Sequence[Unit] | None:
    if quantity.units == u.dimensionless:
        return None
    for family in UNIT_FAMILIES:
        if (1 * family[0]).is_compatible_with(u.dimensionless):
            if any(quantity.units == unit for unit in family):
                return family
        elif quantity.is_compatible_with(family[0]):
            return family
    return None


def _get_representative_magnitude(quantity: Quantity) -> float:
    magnitude = quantity.magnitude
    if isinstance(magnitude, np.ndarray):
        return float(np.mean(np.abs(magnitude)))
    return abs(float(magnitude))


def best_display_unit(quantity: Quantity) -> Unit:
    representative = _get_representative_magnitude(quantity)

    family = _get_unit_family(quantity)
    if family is None:
        return quantity.units
    if representative == 0:
        return family[0]

    representative_quantity = representative * quantity.units
    for unit in reversed(family):
        if representative_quantity.to(unit).magnitude >= 1:
            return unit
    return family[0]


def _round_to_sig_figs(value: float, sig_figs: int = 3) -> float:
    if value == 0:
        return 0.0
    digits = sig_figs - int(math.floor(math.log10(abs(value)))) - 1
    return round(value, digits)


def _round_array_to_sig_figs(values: np.ndarray, sig_figs: int = 3) -> np.ndarray:
    rounded = np.zeros_like(values)
    nonzero_mask = values != 0
    if not np.any(nonzero_mask):
        return rounded

    nonzero_values = values[nonzero_mask].astype(np.float64, copy=False)
    digits = sig_figs - np.floor(np.log10(np.abs(nonzero_values))).astype(np.int64) - 1
    scale = np.power(10.0, digits)
    rounded[nonzero_mask] = np.round(nonzero_values * scale) / scale
    return rounded


def format_quantity_for_display(quantity: Quantity, sig_figs: int = 3) -> Quantity:
    display_unit = best_display_unit(quantity)
    converted = quantity.to(display_unit)
    magnitude = converted.magnitude
    if isinstance(magnitude, np.ndarray):
        rounded = _round_array_to_sig_figs(magnitude, sig_figs)
    else:
        rounded = _round_to_sig_figs(float(magnitude), sig_figs)
    return rounded * display_unit


def human_readable_unit(unit: Unit) -> str:
    output = f"{unit:~P}"
    for special_dimensionless_unit in ["occurrence", "concurrent"]:
        if special_dimensionless_unit in output:
            output = output.replace(special_dimensionless_unit, "")
            if output == "G":
                output = "B"
            break
    return output


def format_display_number(value: float) -> str:
    return np.format_float_positional(value, trim="-")

def display_quantity_as_str(quantity: Quantity, sig_figs: int = 3) -> str:
    formatted_quantity = format_quantity_for_display(quantity, sig_figs)
    return f"{format_display_number(formatted_quantity.magnitude)} {human_readable_unit(formatted_quantity.units)}"



from dataclasses import dataclass
from typing import Any

from pint import Quantity

from efootprint.constants.units import u


@dataclass(frozen=True)
class SankeyGraphSnapshot:
    node_keys_by_index: dict[int, Any]
    full_node_labels: list[str]
    node_color_keys: list[str]
    node_objects: dict[int, Any]
    links: list[tuple[int, int, Quantity]]
    node_total_values: list[Quantity]


class SankeyGraph:
    def __init__(self, truncate_label) -> None:
        self._truncate_label = truncate_label
        self.reset()

    def reset(self) -> None:
        self.node_labels: list[str] = []
        self.full_node_labels: list[str] = []
        self.node_indices: dict[Any, int] = {}
        self.node_color_keys: list[str] = []
        self.node_objects: dict[int, Any] = {}
        self.link_sources: list[int] = []
        self.link_targets: list[int] = []
        self.link_values: list[Quantity] = []
        self.link_index_by_edge: dict[tuple[int, int], int] = {}
        self.node_total_values: list[Quantity] = []

    def add_node(self, label: str, key: Any, color_key: str | None = None, obj: Any = None) -> int:
        if key in self.node_indices:
            return self.node_indices[key]
        idx = len(self.node_labels)
        self.node_labels.append(self._truncate_label(label))
        self.full_node_labels.append(label)
        self.node_indices[key] = idx
        self.node_color_keys.append(color_key or label)
        self.node_total_values.append(0 * u.kg)
        if obj is not None:
            self.node_objects[idx] = obj
        return idx

    def add_link(self, source: int, target: int, value: Quantity) -> None:
        if value.magnitude <= 0:
            return
        edge = (source, target)
        existing_link_idx = self.link_index_by_edge.get(edge)
        if existing_link_idx is None:
            self.link_index_by_edge[edge] = len(self.link_sources)
            self.link_sources.append(source)
            self.link_targets.append(target)
            self.link_values.append(value)
        else:
            self.link_values[existing_link_idx] += value
        self.node_total_values[target] += value

    def snapshot(self) -> SankeyGraphSnapshot:
        return SankeyGraphSnapshot(
            node_keys_by_index={idx: key for key, idx in self.node_indices.items()},
            full_node_labels=list(self.full_node_labels),
            node_color_keys=list(self.node_color_keys),
            node_objects=dict(self.node_objects),
            links=list(zip(self.link_sources, self.link_targets, self.link_values)),
            node_total_values=list(self.node_total_values),
        )

    def reset_links_preserving_root_totals(self) -> list[tuple[int, int, Quantity]]:
        original_links = list(zip(self.link_sources, self.link_targets, self.link_values))
        incoming_nodes = set(self.link_targets)
        root_total_values = {
            idx: self.node_total_values[idx]
            for idx in range(len(self.node_labels))
            if idx not in incoming_nodes and self.node_total_values[idx].magnitude > 0
        }
        self.link_sources = []
        self.link_targets = []
        self.link_values = []
        self.link_index_by_edge = {}
        self.node_total_values = [0 * u.kg for _ in self.node_labels]
        for idx, value in root_total_values.items():
            self.node_total_values[idx] = value
        return original_links



from time import perf_counter

from functools import lru_cache
from inspect import signature

from efootprint.logger import logger


@lru_cache(maxsize=None)
def get_init_signature_params(cls):
    return signature(cls.__init__).parameters


def round_dict(my_dict, round_level):
    for key in my_dict:
        my_dict[key] = round(my_dict[key], round_level)

    return my_dict


def time_it(func):
    def wrapper(*args, **kwargs):
        start_time = perf_counter()
        result = func(*args, **kwargs)
        end_time = perf_counter()
        diff = end_time - start_time
        if diff > 0.000001:
            logger.info(f"Function {func.__name__} took {diff*1000:.1f} ms to execute.")
        return result
    return wrapper
